@@ -11,23 +11,32 @@ def resultset_functional_batchgenerator(cursor, size=100):
 			yield rec
 """
 # =======================================================================
-__version__ = "$Revision: 1.127 $"
 __author__  = "K.Hilbert <Karsten.Hilbert@gmx.net>"
 __license__ = 'GPL v2 or later (details at http://www.gnu.org)'
 
-### imports ###
 # stdlib
-import time, locale, sys, re as regex, os, codecs, types, datetime as pydt, logging
+import time
+import sys
+import os
+import codecs
+import types
+import logging
+import datetime as pydt
+import re as regex
 
 
 # GNUmed
 if __name__ == '__main__':
 	sys.path.insert(0, '../../')
-from Gnumed.pycommon import gmLoginInfo, gmExceptions, gmDateTime, gmBorg, gmI18N, gmLog2
+from Gnumed.pycommon import gmLoginInfo
+from Gnumed.pycommon import gmExceptions
+from Gnumed.pycommon import gmDateTime
+from Gnumed.pycommon import gmBorg
+from Gnumed.pycommon import gmI18N
+from Gnumed.pycommon import gmLog2
 from Gnumed.pycommon.gmTools import prompted_input, u_replacement_character
 
 _log = logging.getLogger('gm.db')
-_log.info(__version__)
 
 
 # 3rd party
@@ -37,7 +46,6 @@ except ImportError:
 	_log.exception("Python database adapter psycopg2 not found.")
 	print "CRITICAL ERROR: Cannot find module psycopg2 for connecting to the database server."
 	raise
-### imports ###
 
 
 _log.info('psycopg2 version: %s' % dbapi.__version__)
@@ -105,7 +113,11 @@ known_schema_hashes = {
 	13: 'fab7c1ae408a6530c47f9b5111a0841e',
 	14: 'e170d543f067d1ea60bfe9076b1560cf',
 	15: '70012ff960b77ecdff4981c94b5b55b6',
-	16: '0bcf44ca22c479b52976e5eda1de8161'
+	16: '0bcf44ca22c479b52976e5eda1de8161',
+	17: '161428ee97a00e3bf56168c3a15b7b50',
+	18: 'a0f9efcabdecfb4ddb6d8c0b69c02092',
+	#19: '419e5225259c53dd36ad80d82066ff02'	# 19.0 only
+	19: '9765373098b03fb208332498f34cd4b5'
 }
 
 map_schema_hash2version = {
@@ -123,7 +135,11 @@ map_schema_hash2version = {
 	'fab7c1ae408a6530c47f9b5111a0841e': 13,
 	'e170d543f067d1ea60bfe9076b1560cf': 14,
 	'70012ff960b77ecdff4981c94b5b55b6': 15,
-	'0bcf44ca22c479b52976e5eda1de8161': 16
+	'0bcf44ca22c479b52976e5eda1de8161': 16,
+	'161428ee97a00e3bf56168c3a15b7b50': 17,
+	'a0f9efcabdecfb4ddb6d8c0b69c02092': 18,
+	#'419e5225259c53dd36ad80d82066ff02': 19	# 19.0. only
+	'9765373098b03fb208332498f34cd4b5': 19
 }
 
 map_client_branch2required_db_version = {
@@ -135,8 +151,11 @@ map_client_branch2required_db_version = {
 	u'0.7': 13,
 	u'0.8': 14,
 	u'0.9': 15,
-	u'1.0': 16,
-	u'1.1': 16
+	u'1.0': 16,		# intentional duplicate with 1.1
+	u'1.1': 16,
+	u'1.2': 17,
+	u'1.3': 18,
+	u'1.4': 19
 }
 
 # get columns and data types for a given table
@@ -163,6 +182,28 @@ where
 order by
 	cols.ordinal_position"""
 
+# only works for single-column FKs but that's fine
+# needs gm-dbo, any-doc won't work
+SQL_foreign_key_name = u"""SELECT tc.constraint_schema, tc.constraint_name
+FROM
+	information_schema.table_constraints tc
+		INNER JOIN information_schema.constraint_column_usage ccu USING (constraint_catalog, constraint_schema, constraint_name)
+			INNER JOIN information_schema.key_column_usage kcu USING (constraint_catalog, constraint_schema, constraint_name)
+WHERE
+	tc.constraint_type = 'FOREIGN KEY'
+		AND
+	kcu.table_schema = %(src_schema)s
+		AND
+	kcu.table_name = %(src_tbl)s
+		AND
+	kcu.column_name = %(src_col)s
+		AND
+	ccu.table_schema = %(target_schema)s
+		AND
+	ccu.table_name = %(target_tbl)s
+		AND
+	ccu.column_name = %(target_col)s"""
+
 # =======================================================================
 # module globals API
 # =======================================================================
@@ -319,7 +360,7 @@ def __request_login_params_tui():
 	print "\nPlease enter the required login parameters:"
 	try:
 		login.host = prompted_input(prompt = "host ('' = non-TCP/IP)", default = '')
-		login.database = prompted_input(prompt = "database", default = 'gnumed_v16')
+		login.database = prompted_input(prompt = "database", default = 'gnumed_v19')
 		login.user = prompted_input(prompt = "user name", default = '')
 		tmp = 'password for "%s" (not shown): ' % login.user
 		login.password = getpass.getpass(tmp)
@@ -462,10 +503,10 @@ def database_schema_compatible(link_obj=None, version=None, verbose=True):
 		_log.error('detected: %s (%s)' % (get_schema_version(link_obj=link_obj), rows[0]['md5']))
 		if verbose:
 			_log.debug('schema dump follows:')
-			for line in get_schema_structure(link_obj=link_obj).split():
+			for line in get_schema_structure(link_obj = link_obj).split():
 				_log.debug(line)
 			_log.debug('schema revision history dump follows:')
-			for line in get_schema_revision_history(link_obj=link_obj):
+			for line in get_schema_revision_history(link_obj = link_obj):
 				_log.debug(u' - '.join(line))
 		return False
 	_log.info('detected schema version [%s], hash [%s]' % (map_schema_hash2version[rows[0]['md5']], rows[0]['md5']))
@@ -487,15 +528,27 @@ def get_schema_hash(link_obj=None):
 	return rows[0]['md5']
 #------------------------------------------------------------------------
 def get_schema_revision_history(link_obj=None):
-	cmd = u"""
-select
-	imported::text,
-	version,
-	filename
-from gm.schema_revision
-order by imported
-"""
-	rows, idx = run_ro_queries(link_obj=link_obj, queries = [{'cmd': cmd}])
+
+	if table_exists(link_obj = link_obj, schema = 'gm', table = 'schema_revision'):
+		cmd = u"""
+			SELECT
+				imported::text,
+				version,
+				filename
+			FROM gm.schema_revision
+			ORDER BY imported"""
+	elif table_exists(link_obj = link_obj, schema = 'public', table = 'gm_schema_revision'):
+		cmd = u"""
+			SELECT
+				imported::text,
+				version,
+				filename
+			FROM public.gm_schema_revision
+			ORDER BY imported"""
+	else:
+		return []
+
+	rows, idx = run_ro_queries(link_obj = link_obj, queries = [{'cmd': cmd}])
 	return rows
 #------------------------------------------------------------------------
 def get_current_user():
@@ -553,6 +606,27 @@ where
 	)
 
 	return rows
+
+#------------------------------------------------------------------------
+def get_foreign_key_names(src_schema=None, src_table=None, src_column=None, target_schema=None, target_table=None, target_column=None, link_obj=None):
+
+	args = {
+		'src_schema': src_schema,
+		'src_tbl': src_table,
+		'src_col': src_column,
+		'target_schema': target_schema,
+		'target_tbl': target_table,
+		'target_col': target_column
+	}
+
+	rows, idx = run_ro_queries (
+		link_obj = link_obj,
+		queries = [{'cmd': SQL_foreign_key_name, 'args': args}],
+		get_col_idx = False
+	)
+
+	return rows
+
 #------------------------------------------------------------------------
 def get_child_tables(schema='public', table=None, link_obj=None):
 	"""Return child tables of <table>."""
@@ -644,16 +718,17 @@ def export_translations_from_database(filename=None):
 	tx_file = codecs.open(filename, 'wb', 'utf8')
 	tx_file.write(u'-- GNUmed database string translations exported %s\n' % gmDateTime.pydt_now_here().strftime('%Y-%m-%d %H:%M'))
 	tx_file.write(u'-- - contains translations for each of [%s]\n' % u', '.join(get_translation_languages()))
-	tx_file.write(u'-- - user language is set to [%s]\n\n' % get_current_user_language())
+	tx_file.write(u'-- - user database language is set to [%s]\n\n' % get_current_user_language())
 	tx_file.write(u'-- Please email this file to <gnumed-devel@gnu.org>.\n')
 	tx_file.write(u'-- ----------------------------------------------------------------------------------------------\n\n')
-	tx_file.write(u'set default_transaction_read_only to off\n\n')
+	tx_file.write(u'set default_transaction_read_only to off;\n\n')
+	tx_file.write(u"set client_encoding to 'utf-8';\n\n")
 	tx_file.write(u'\\unset ON_ERROR_STOP\n\n')
 
 	cmd = u'SELECT lang, orig, trans FROM i18n.translations ORDER BY lang, orig'
 	rows, idx = run_ro_queries(queries = [{'cmd': cmd}], get_col_idx = False)
 	for row in rows:
-		line = u"select i18n.upd_tx(quote_literal(E'%s'), quote_literal(E'%s'), quote_literal(E'%s'));\n" % (
+		line = u"select i18n.upd_tx(E'%s', E'%s', E'%s');\n" % (
 			row['lang'].replace("'", "\\'"),
 			row['orig'].replace("'", "\\'"),
 			row['trans'].replace("'", "\\'")
@@ -810,125 +885,16 @@ def force_user_language(language=None):
 		'cmd': u'select i18n.force_curr_lang(%(lang)s)',
 		'args': {'lang': language}
 	}])
-#------------------------------------------------------------------------
-#------------------------------------------------------------------------
-text_expansion_keywords = None
 
-def get_text_expansion_keywords():
-	global text_expansion_keywords
-	if text_expansion_keywords is not None:
-		return text_expansion_keywords
-
-	cmd = u"""select keyword, public_expansion, private_expansion, owner from clin.v_keyword_expansions"""
-	rows, idx = run_ro_queries(queries = [{'cmd': cmd}])
-	text_expansion_keywords = rows
-
-	_log.info('retrieved %s text expansion keywords', len(text_expansion_keywords))
-
-	return text_expansion_keywords
-#------------------------------------------------------------------------
-def expand_keyword(keyword = None):
-
-	# Easter Egg ;-)
-	if keyword == u'$$steffi':
-		return u'Hai, play !  Versucht das ! (Keks dazu ?)  :-)'
-
-	cmd = u"""select expansion from clin.v_your_keyword_expansions where keyword = %(kwd)s"""
-	rows, idx = run_ro_queries(queries = [{'cmd': cmd, 'args': {'kwd': keyword}}])
-
-	if len(rows) == 0:
-		return None
-
-	return rows[0]['expansion']
-#------------------------------------------------------------------------
-def get_keyword_expansion_candidates(keyword = None):
-
-	if keyword is None:
-		return []
-
-	get_text_expansion_keywords()
-
-	candidates = []
-	for kwd in text_expansion_keywords:
-		if kwd['keyword'].startswith(keyword):
-			candidates.append(kwd['keyword'])
-
-	return candidates
-#------------------------------------------------------------------------
-def add_text_expansion(keyword=None, expansion=None, public=None):
-
-	if public:
-		cmd = u"SELECT 1 from clin.v_keyword_expansions where public_expansion is true and keyword = %(kwd)s"
-	else:
-		cmd = u"SELECT 1 from clin.v_your_keyword_expansions where private_expansion is true and keyword = %(kwd)s"
-
-	rows, idx = run_ro_queries(queries = [{'cmd': cmd, 'args': {'kwd': keyword}}])
-	if len(rows) != 0:
-		return False
-
-	if public:
-		cmd = u"""
-insert into clin.keyword_expansion (keyword, expansion, fk_staff)
-values (%(kwd)s, %(exp)s, null)"""
-	else:
-		cmd = u"""
-insert into clin.keyword_expansion (keyword, expansion, fk_staff)
-values (%(kwd)s, %(exp)s, (SELECT pk from dem.staff where db_user = current_user))"""
-
-	rows, idx = run_rw_queries(queries = [{'cmd': cmd, 'args': {'kwd': keyword, 'exp': expansion}}])
-
-	global text_expansion_keywords
-	text_expansion_keywords = None
-
-	return True
-#------------------------------------------------------------------------
-def delete_text_expansion(keyword):
-	cmd = u"""
-delete from clin.keyword_expansion where
-	keyword = %(kwd)s and (
-		(fk_staff = (SELECT pk from dem.staff where db_user = current_user))
-			or
-		(fk_staff is null and owner = current_user)
-	)"""
-	rows, idx = run_rw_queries(queries = [{'cmd': cmd, 'args': {'kwd': keyword}}])
-
-	global text_expansion_keywords
-	text_expansion_keywords = None
-#------------------------------------------------------------------------
-def edit_text_expansion(keyword, expansion):
-
-	cmd1 = u"""
-		DELETE FROM clin.keyword_expansion
-		WHERE
-			keyword = %(kwd)s
-				AND
-			fk_staff = (SELECT pk FROM dem.staff WHERE db_user = current_user)"""
-
-	cmd2 = u"""
-		INSERT INTO clin.keyword_expansion (
-			keyword, expansion, fk_staff
-		) VALUES (
-			%(kwd)s,
-			%(exp)s,
-			(SELECT pk FROM dem.staff WHERE db_user = current_user)
-		)"""
-	args = {'kwd': keyword, 'exp': expansion}
-	rows, idx = run_rw_queries(queries = [
-		{'cmd': cmd1, 'args': args},
-		{'cmd': cmd2, 'args': args},
-	])
-
-	global text_expansion_keywords
-	text_expansion_keywords = None
 # =======================================================================
 # query runners and helpers
 # =======================================================================
 def send_maintenance_notification():
-	cmd = u'notify "db_maintenance_warning:"'
+	cmd = u'notify "db_maintenance_warning"'
 	run_rw_queries(queries = [{'cmd': cmd}], return_data = False)
 #------------------------------------------------------------------------
 def send_maintenance_shutdown():
-	cmd = u'notify "db_maintenance_disconnect:"'
+	cmd = u'notify "db_maintenance_disconnect"'
 	run_rw_queries(queries = [{'cmd': cmd}], return_data = False)
 #------------------------------------------------------------------------
 def is_pg_interval(candidate=None):
@@ -943,6 +909,67 @@ def is_pg_interval(candidate=None):
 			return True
 		except:
 			return False
+
+#------------------------------------------------------------------------
+def lock_row(link_obj=None, table=None, pk=None, exclusive=False):
+	"""Uses pg_advisory(_shared).
+
+	- locks stack upon each other and need one unlock per lock
+	- same connection:
+		- all locks succeed
+	- different connections:
+		- shared + shared succeed
+		- shared + exclusive fail
+	"""
+	_log.debug('locking row: [%s] [%s] (exclusive: %s)', table, pk, exclusive)
+	if exclusive:
+		cmd = u"""SELECT pg_try_advisory_lock('%s'::regclass::oid::int, %s)""" % (table, pk)
+	else:
+		cmd = u"""SELECT pg_try_advisory_lock_shared('%s'::regclass::oid::int, %s)""" % (table, pk)
+	rows, idx = run_ro_queries(link_obj = link_obj, queries = [{'cmd': cmd}], get_col_idx = False)
+	if rows[0][0]:
+		return True
+	_log.warning('cannot lock row: [%s] [%s] (exclusive: %s)', table, pk, exclusive)
+	return False
+
+#------------------------------------------------------------------------
+def unlock_row(link_obj=None, table=None, pk=None, exclusive=False):
+	"""Uses pg_advisory_unlock(_shared).
+
+	- each lock needs one unlock
+	"""
+	_log.debug('trying to unlock row: [%s] [%s] (exclusive: %s)', table, pk, exclusive)
+	if exclusive:
+		cmd = u"SELECT pg_advisory_unlock('%s'::regclass::oid::int, %s)" % (table, pk)
+	else:
+		cmd = u"SELECT pg_advisory_unlock_shared('%s'::regclass::oid::int, %s)" % (table, pk)
+	rows, idx = run_ro_queries(link_obj = link_obj, queries = [{'cmd': cmd}], get_col_idx = False)
+	if rows[0][0]:
+		return True
+	_log.warning('cannot unlock row: [%s] [%s] (exclusive: %s)', table, pk, exclusive)
+	return False
+
+#------------------------------------------------------------------------
+def row_is_locked(table=None, pk=None):
+	"""Looks at pk_locks
+
+	- does not take into account locks other than 'advisory', however
+	"""
+	cmd = u"""SELECT EXISTS (
+		SELECT 1 FROM pg_locks WHERE
+			classid = '%s'::regclass::oid::int
+				AND
+			objid = %s
+				AND
+			locktype = 'advisory'
+	)""" % (table, pk)
+	rows, idx = run_ro_queries(queries = [{'cmd': cmd}], get_col_idx = False)
+	if rows[0][0]:
+		_log.debug('row is locked: [%s] [%s]', table, pk)
+		return True
+	_log.debug('row is NOT locked: [%s] [%s]', table, pk)
+	return False
+
 #------------------------------------------------------------------------
 def bytea2file(data_query=None, filename=None, chunk_size=0, data_size=None, data_size_query=None):
 	outfile = file(filename, 'wb')
@@ -991,28 +1018,26 @@ def bytea2file_object(data_query=None, file_obj=None, chunk_size=0, data_size=No
 			conn.rollback()
 			return True
 
-	_log.debug('expecting bytea data of size: [%s] bytes' % data_size)
-	_log.debug('using chunk size of: [%s] bytes' % chunk_size)
-
-	# chunk size of 0 means "retrieve whole field at once"
+	max_chunk_size = 1024 * 1024 * 20			# 20 MB, works for typical CR DICOMs
 	if chunk_size == 0:
-		chunk_size = data_size
-		_log.debug('chunk size [0] bytes: retrieving all data at once')
+		chunk_size = min(data_size, max_chunk_size)
+
+	_log.debug('expecting %s bytes of BYTEA data in chunks of %s bytes', data_size, chunk_size)
 
 	# Windoze sucks: it can't transfer objects of arbitrary size,
 	# anyways, we need to split the transfer,
 	# however, only possible if postgres >= 7.2
 	needed_chunks, remainder = divmod(data_size, chunk_size)
-	_log.debug('chunks to retrieve: [%s]' % needed_chunks)
-	_log.debug('remainder to retrieve: [%s] bytes' % remainder)
+	_log.debug('# of chunks: %s; remainder: %s bytes', needed_chunks, remainder)
 
-	# try setting "bytea_output"
-	# - fails if not necessary
-	# - succeeds if necessary
-	try:
-		run_ro_queries(link_obj = conn, queries = [{'cmd': u"set bytea_output to 'escape'"}])
-	except dbapi.ProgrammingError:
-		_log.debug('failed to set bytea_output to "escape", not necessary')
+#	# since we now require PG 9.1 we can disable this workaround:
+#	# try setting "bytea_output"
+#	# - fails if not necessary
+#	# - succeeds if necessary
+#	try:
+#		run_ro_queries(link_obj = conn, queries = [{'cmd': u"set bytea_output to 'escape'"}])
+#	except dbapi.ProgrammingError:
+#		_log.debug('failed to set bytea_output to "escape", not necessary')
 
 	# retrieve chunks, skipped if data size < chunk size,
 	# does this not carry the danger of cutting up multi-byte escape sequences ?
@@ -1050,14 +1075,19 @@ def bytea2file_object(data_query=None, file_obj=None, chunk_size=0, data_size=No
 	conn.rollback()
 	return True
 #------------------------------------------------------------------------
-def file2bytea(query=None, filename=None, args=None, conn=None):
+def file2bytea(query=None, filename=None, args=None, conn=None, file_md5=None):
 	"""Store data from a file into a bytea field.
 
 	The query must:
 	- be in unicode
 	- contain a format spec identifying the row (eg a primary key)
 	  matching <args> if it is an UPDATE
-	- contain a format spec %(data)s::bytea
+	- contain a format spec " <field> = %(data)s::bytea"
+
+	The query CAN return the MD5 of the inserted data:
+		RETURNING md5(<field>) AS md5
+	in which case it will compare it to the md5
+	of the file.
 	"""
 	# read data from file
 	infile = file(filename, "rb")
@@ -1075,12 +1105,25 @@ def file2bytea(query=None, filename=None, args=None, conn=None):
 	else:
 		close_conn = False
 
-	run_rw_queries(link_obj = conn, queries = [{'cmd': query, 'args': args}], end_tx = True)
+	rows, idx = run_rw_queries(link_obj = conn, queries = [{'cmd': query, 'args': args}], end_tx = False, return_data = (file_md5 is not None))
+
+	success_status = True
+	if file_md5 is None:
+		conn.commit()
+	else:
+		db_md5 = rows[0]['md5']
+		if file_md5 != db_md5:
+			conn.rollback()
+			success_status = False
+			_log.error('MD5 sums of data file and database BYTEA field do not match: [file::%s] <> [DB::%s]', file_md5, db_md5)
+		else:
+			conn.commit()
+			_log.debug('MD5 sums of data file and database BYTEA field match: [file::%s] = [DB::%s]', file_md5, db_md5)
 
 	if close_conn:
 		conn.close()
 
-	return
+	return success_status
 #------------------------------------------------------------------------
 def sanitize_pg_regex(expression=None, escape_all=False):
 	"""Escape input for use in a PostgreSQL regular expression.
@@ -1149,17 +1192,49 @@ def run_ro_queries(link_obj=None, queries=None, verbose=False, return_data=True,
 			curs.execute(query['cmd'], args)
 			if verbose:
 				_log.debug('ran query: [%s]', curs.query)
-				_log.debug('PG status message: %s', curs.statusmessage)
+				if curs.statusmessage != u'':
+					_log.debug('PG status message: %s', curs.statusmessage)
 				_log.debug('cursor description: %s', str(curs.description))
-		except:
-			# FIXME: use .pgcode
+		except dbapi.Error as pg_exc:
+			_log.error('query failed: [%s]', curs.query)
+			if curs.statusmessage != u'':
+				_log.error('PG status message: %s', curs.statusmessage)
+			pg_exc = make_pg_exception_fields_unicode(pg_exc)
+			_log.error('PG error code: %s', pg_exc.pgcode)
+			if pg_exc.pgerror is not None:
+				_log.error(u'PG error message: %s', pg_exc.u_pgerror)
 			try:
 				curs_close()
 			except dbapi.InterfaceError:
 				_log.exception('cannot close cursor')
 			tx_rollback()		# need to rollback so ABORT state isn't preserved in pooled conns
+			if pg_exc.pgcode == sql_error_codes.INSUFFICIENT_PRIVILEGE:
+				details = u'Query: [%s]' % curs.query.strip().strip(u'\n').strip().strip(u'\n')
+				if curs.statusmessage != u'':
+					details = u'Status: %s\n%s' % (
+						curs.statusmessage.strip().strip(u'\n').strip().strip(u'\n'),
+						details
+					)
+				if pg_exc.pgerror is None:
+					msg = u'[%s]' % pg_exc.pgcode
+				else:
+					msg = u'[%s]: %s' % (pg_exc.pgcode, pg_exc.u_pgerror)
+				raise gmExceptions.AccessDenied (
+					msg,
+					source = u'PostgreSQL',
+					code = pg_exc.pgcode,
+					details = details
+				)
+			raise
+		except:
 			_log.error('query failed: [%s]', curs.query)
-			_log.error('PG status message: %s', curs.statusmessage)
+			if curs.statusmessage != u'':
+				_log.error('PG status message: %s', curs.statusmessage)
+			try:
+				curs_close()
+			except dbapi.InterfaceError:
+				_log.exception('cannot close cursor')
+			tx_rollback()		# need to rollback so ABORT state isn't preserved in pooled conns
 			raise
 
 	data = None
@@ -1225,24 +1300,24 @@ def run_rw_queries(link_obj=None, queries=None, end_tx=False, return_data=None,
 	if isinstance(link_obj, dbapi._psycopg.cursor):
 		conn_close = __noop
 		conn_commit = __noop
-		conn_rollback = __noop
+		tx_rollback = __noop
 		curs = link_obj
 		curs_close = __noop
 	elif isinstance(link_obj, dbapi._psycopg.connection):
 		conn_close = __noop
 		if end_tx:
 			conn_commit = link_obj.commit
-			conn_rollback = link_obj.rollback
+			tx_rollback = link_obj.rollback
 		else:
 			conn_commit = __noop
-			conn_rollback = __noop
+			tx_rollback = __noop
 		curs = link_obj.cursor()
 		curs_close = curs.close
 	elif link_obj is None:
 		conn = get_connection(readonly=False)
 		conn_close = conn.close
 		conn_commit = conn.commit
-		conn_rollback = conn.rollback
+		tx_rollback = conn.rollback
 		curs = conn.cursor()
 		curs_close = curs.close
 	else:
@@ -1250,7 +1325,7 @@ def run_rw_queries(link_obj=None, queries=None, end_tx=False, return_data=None,
 
 	for query in queries:
 		if type(query['cmd']) is not types.UnicodeType:
-			print "run_rw_queries(): non-unicode query"
+			print "gmPG2.run_rw_queries(): non-unicode query"
 			print query['cmd']
 		try:
 			args = query['args']
@@ -1258,12 +1333,44 @@ def run_rw_queries(link_obj=None, queries=None, end_tx=False, return_data=None,
 			args = None
 		try:
 			curs.execute(query['cmd'], args)
+		except dbapi.Error as pg_exc:
+			_log.error('RW query failed: [%s]', curs.query)
+			if curs.statusmessage != u'':
+				_log.error('PG status message: %s', curs.statusmessage)
+			pg_exc = make_pg_exception_fields_unicode(pg_exc)
+			_log.error(u'PG error code: %s', pg_exc.pgcode)
+			if pg_exc.pgerror is not None:
+				_log.error(u'PG error message: %s', pg_exc.u_pgerror)
+			try:
+				curs_close()
+				tx_rollback()			# just for good measure
+				conn_close()
+			except dbapi.InterfaceError:
+				_log.exception('cannot cleanup')
+			if pg_exc.pgcode == sql_error_codes.INSUFFICIENT_PRIVILEGE:
+				details = u'Query: [%s]' % curs.query.strip().strip(u'\n').strip().strip(u'\n')
+				if curs.statusmessage != u'':
+					details = u'Status: %s\n%s' % (
+						curs.statusmessage.strip().strip(u'\n').strip().strip(u'\n'),
+						details
+					)
+				if pg_exc.pgerror is None:
+					msg = u'[%s]' % pg_exc.pgcode
+				else:
+					msg = u'[%s]: %s' % (pg_exc.pgcode, pg_exc.u_pgerror)
+				raise gmExceptions.AccessDenied (
+					msg,
+					source = u'PostgreSQL',
+					code = pg_exc.pgcode,
+					details = details
+				)
+			raise
 		except:
 			_log.exception('error running RW query')
 			gmLog2.log_stack_trace()
 			try:
 				curs_close()
-				conn_rollback()
+				tx_rollback()
 				conn_close()
 			except dbapi.InterfaceError:
 				_log.exception('cannot cleanup')
@@ -1280,7 +1387,7 @@ def run_rw_queries(link_obj=None, queries=None, end_tx=False, return_data=None,
 			gmLog2.log_stack_trace()
 			try:
 				curs_close()
-				conn_rollback()
+				tx_rollback()
 				conn_close()
 			except dbapi.InterfaceError:
 				_log.exception('cannot cleanup')
@@ -1503,6 +1610,7 @@ def get_connection(dsn=None, readonly=True, encoding=None, verbose=False, pooled
 
 	# - transaction isolation level
 	if readonly:
+		# alter-database default, checked at connect, no need to set now
 		iso_level = u'read committed'
 	else:
 		conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE)
@@ -1517,22 +1625,22 @@ def get_connection(dsn=None, readonly=True, encoding=None, verbose=False, pooled
 
 	conn.commit()
 
-	# FIXME: remove this whole affair once either 9.0 is standard (Ubuntu 10 LTS is
-	# FIXME: PG 8.4, however!) or else when psycopg2 supports a workaround
-	#
-	# - bytea data format
-	# PG 9.0 switched to - by default - using "hex" rather than "escape",
-	# however, psycopg2's linked with a pre-9.0 libpq do assume "escape"
-	# as the transmission mode for bytea output,
-	# so try to set this setting back to "escape",
-	# if that's not possible the reason will be that PG < 9.0 does not support
-	# that setting - which also means we don't need it and can ignore the
-	# failure
-	cmd = "set bytea_output to 'escape'"
-	try:
-		curs.execute(cmd)
-	except dbapi.ProgrammingError:
-		_log.error('cannot set bytea_output format')
+#	# FIXME: remove this whole affair once either 9.0 is standard (Ubuntu 10 LTS is
+#	# FIXME: PG 8.4, however!) or else when psycopg2 supports a workaround
+#	#
+#	# - bytea data format
+#	# PG 9.0 switched to - by default - using "hex" rather than "escape",
+#	# however, psycopg2's linked with a pre-9.0 libpq do assume "escape"
+#	# as the transmission mode for bytea output,
+#	# so try to set this setting back to "escape",
+#	# if that's not possible the reason will be that PG < 9.0 does not support
+#	# that setting - which also means we don't need it and can ignore the
+#	# failure
+#	cmd = "set bytea_output to 'escape'"
+#	try:
+#		curs.execute(cmd)
+#	except dbapi.ProgrammingError:
+#		_log.error('cannot set bytea_output format')
 
 	curs.close()
 	conn.commit()
@@ -1638,9 +1746,10 @@ def sanity_check_database_settings():
 		u'full_page_writes': [u'on', u'data loss/corruption', False],
 		u'lc_messages': [u'C', u'suboptimal error detection', False],
 		u'password_encryption': [u'on', u'breach of confidentiality', False],
-		u'regex_flavor': [u'advanced', u'query breakage', False],					# 9.0 doesn't support this anymore, default now advanced anyway
+		#u'regex_flavor': [u'advanced', u'query breakage', False],					# 9.0 doesn't support this anymore, default now advanced anyway
 		u'synchronous_commit': [u'on', u'data loss/corruption', False],
-		u'sql_inheritance': [u'on', u'query breakage, data loss/corruption', True]
+		u'sql_inheritance': [u'on', u'query breakage, data loss/corruption', True],
+		u'ignore_checksum_failure': [u'off', u'data loss/corruption', False]		# starting with PG 9.3
 	}
 
 	from Gnumed.pycommon import gmCfg2
@@ -1706,7 +1815,26 @@ def __log_PG_settings(curs=None):
 	for setting in settings:
 		_log.debug(u'PG option [%s]: %s', setting[0], setting[1])
 	return True
-# =======================================================================
+#========================================================================
+def make_pg_exception_fields_unicode(exc):
+
+	if not isinstance(exc, dbapi.Error):
+		return exc
+
+	if exc.pgerror is None:
+		try:
+			msg = exc.args[0]
+		except (AttributeError, IndexError, TypeError):
+			return exc
+		# assumption
+		exc.u_pgerror = unicode(msg, gmI18N.get_encoding(), 'replace')
+		return exc
+
+	# assumption
+	exc.u_pgerror = unicode(exc.pgerror, gmI18N.get_encoding(), 'replace').strip().strip(u'\n').strip().strip(u'\n')
+
+	return exc
+#------------------------------------------------------------------------
 def extract_msg_from_pg_exception(exc=None):
 
 	try:
@@ -1714,6 +1842,7 @@ def extract_msg_from_pg_exception(exc=None):
 	except (AttributeError, IndexError, TypeError):
 		return u'cannot extract message from exception'
 
+	# assumption
 	return unicode(msg, gmI18N.get_encoding(), 'replace')
 # =======================================================================
 class cAuthenticationError(dbapi.OperationalError):
@@ -1796,13 +1925,6 @@ psycopg2.extensions.register_type(psycopg2._psycopg.UNICODEARRAY)
 # check in 0.9:
 psycopg2.extensions.register_adapter(pydt.datetime, cAdapterPyDateTime)
 
-## remove for 0.9
-#try:
-#	import mx.DateTime as mxDT
-##	psycopg2.extensions.register_adapter(mxDT.DateTimeType, cAdapterMxDateTime)
-#except ImportError:
-#	_log.warning('cannot import mx.DateTime')
-
 # do NOT adapt *lists* to "... IN (*) ..." syntax because we want
 # them adapted to "... ARRAY[]..." so we can support PG arrays
 
@@ -2068,14 +2190,6 @@ if __name__ == "__main__":
 	def test_sanity_check_time_skew():
 		sanity_check_time_skew()
 	#--------------------------------------------------------------------
-	def test_keyword_expansion():
-		print "keywords, from database:"
-		print get_text_expansion_keywords()
-		print "keywords, cached:"
-		print get_text_expansion_keywords()
-		print "'$keyword' expands to:"
-		print expand_keyword(keyword = u'$dvt')
-	#--------------------------------------------------------------------
 	def test_get_foreign_key_details():
 		for row in get_foreign_keys2column (
 			schema = u'dem',
@@ -2129,11 +2243,70 @@ if __name__ == "__main__":
 		cmd = u"SELECT %(dt)s"
 
 		#cmd = u"SELECT 'infinity'::timestamp with time zone"
+
+		cmd = u"""
+SELECT to_timestamp (foofoo,'YYMMDD.HH24MI') FROM (
+	SELECT REGEXP_REPLACE (
+		't1.130729.0902.tif',			-- string
+		E'(.1)\.([0-9\.]+)(\.tif)',		-- pattern
+		E'\\\\2'						-- replacement
+	) AS foofoo
+) AS foo"""
 		rows, idx = run_ro_queries(queries = [{'cmd': cmd, 'args': args}], get_col_idx = False)
 		print rows
+		print rows[0]
+		print rows[0][0]
 	#--------------------------------------------------------------------
 	def test_schema_exists():
 		print schema_exists()
+	#--------------------------------------------------------------------
+	def test_row_locks():
+		row_is_locked(table = 'dem.identity', pk = 12)
+
+		print "1st connection:"
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+		print " 1st shared lock succeeded:", lock_row(table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+
+		print "   2nd shared lock should succeed:", lock_row(table = 'dem.identity', pk = 12, exclusive = False)
+		print "   `-> unlock succeeded:", unlock_row(table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+		print "   exclusive lock should succeed:", lock_row(table = 'dem.identity', pk = 12, exclusive = True)
+		print "   `-> unlock succeeded:", unlock_row(table = 'dem.identity', pk = 12, exclusive = True)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+
+		print "2nd connection:"
+		conn = get_raw_connection(readonly=True)
+		print " shared lock should succeed:", lock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = False)
+		print " `-> unlock succeeded:", unlock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+		print " exclusive lock succeeded ?", lock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = True), "(should fail)"
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+
+		print "1st connection:"
+		print " unlock succeeded:", unlock_row(table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+
+		print "2nd connection:"
+		print " exclusive lock should succeed", lock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = True)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+		print "  shared lock should succeed:", lock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = False)
+		print "  `-> unlock succeeded:", unlock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+		print " unlock succeeded:", unlock_row(link_obj = conn, table = 'dem.identity', pk = 12, exclusive = False)
+		print " locked:", row_is_locked(table = 'dem.identity', pk = 12)
+
+		conn.close()
+	#--------------------------------------------------------------------
+	def test_get_foreign_key_names():
+		print get_foreign_key_names (
+			src_schema = 'dem',
+			src_table = 'names',
+			src_column = 'id_identity',
+			target_schema = 'dem',
+			target_table = 'identity',
+			target_column = 'pk'
+		)
 
 	#--------------------------------------------------------------------
 	# run tests
@@ -2148,11 +2321,12 @@ if __name__ == "__main__":
 	#test_sanitize_pg_regex()
 	#test_is_pg_interval()
 	#test_sanity_check_time_skew()
-	#test_keyword_expansion()
 	#test_get_foreign_key_details()
 	#test_set_user_language()
 	#test_get_schema_revision_history()
 	#test_run_query()
-	test_schema_exists()
+	#test_schema_exists()
+	#test_get_foreign_key_names()
+	test_row_locks()
 
 # ======================================================================
